import asyncio
from .request import Request
import inspect
from .item import Item
from ..utils import logger
from .exceptions import EngineError

__all__ = ['Engine']

class Engine(object):

    __slots__ = ['pipelines', 'spiders', 'middlewares', 'sched', 'loop', 'timeout', 'headers']

    def __init__(self, loop=None):
        self.pipelines = []
        self.spiders = {}
        self.middlewares = []
        self.sched = None
        self.loop = loop
        if not self.loop:
            self.loop = asyncio.get_event_loop()

        self.timeout = None
        self.headers = {
            'User-Agent': 'Grapy/1.0'
        }

    def set_spiders(self, spiders):
        self.spiders = {}
        if isinstance(spiders, dict):
            for name, spider in spiders.items():
                self.spiders[name] = spider

        else:
            self.add_spiders(spiders)

    def add_spiders(self, spiders):
        for spider in spiders:
            self.add_spider(spider)

    def add_spider(self, spider):
        if spider.name in self.spiders.keys():
            raise EngineError('Spider[%s] is already exists'%spider.name)
        self.spiders[spider.name] = spider

    def remove_spider(self, spider_name):
        self.spiders.pop(spider_name)

    def get_spider(self, name):
        spider = self.spiders.get(name)
        if spider:
            return spider
        else:
            raise EngineError('Spider[%s] is not found'%name)

    def set_pipelines(self, pipelines):
        self.pipelines = pipelines

    def set_middlewares(self, middlewares):
        self.middlewares = middlewares

    def set_sched(self, sched):
        self.sched = sched
        self.sched.engine = self

    @asyncio.coroutine
    def process(self, req):
        req.engine = self
        req = yield from self.process_middleware('before_process_request', req)

        rsp = yield from req.request()

        rsp.req = req

        rsp = yield from self.process_middleware('after_process_response', rsp)

        yield from self.process_response(rsp)

    @asyncio.coroutine
    def process_middleware(self, name, obj):
        for mid in self.middlewares:
            if hasattr(mid, name):
                func = getattr(mid, name)
                obj = func(obj)
                if isinstance(obj, asyncio.Future) or inspect.isgenerator(obj):
                    obj = yield from obj

        return obj

    @asyncio.coroutine
    def process_item(self, item, pipelines=None):
        if not pipelines:
            pipelines = self.pipelines

        for pip in pipelines:

            item = pip.process(item)
            if isinstance(item, asyncio.Future) or inspect.isgenerator(item):
                item = yield from item

    @asyncio.coroutine
    def process_response(self, rsp):
        spider_name = rsp.req.spider
        callback = rsp.req.callback
        args = rsp.req.callback_args
        spider = self.get_spider(spider_name)
        func = getattr(spider, callback)
        items = func(rsp, *args)
        if items is None:
            return
        for item in items:
            if isinstance(item, Request):
                item.spider = spider.name
                logger.debug('Find url[{}] on requset[{}] by spider[{}]'.\
                        format(item.url, rsp.url, spider.name))

                item.group = rsp.req.group
                item.ref = rsp.url

                yield from self.push_req(item)
            elif isinstance(item, Item):
                yield from self.push_item(item)
            else:
                raise EngineError('Unknow type')

    @asyncio.coroutine
    def push_req(self, req, middleware=True):
        if middleware:
            req = yield from self.process_middleware('before_push_request', req)

        req = self.sched.push_req(req)
        if isinstance(req, asyncio.Future) or inspect.isgenerator(req):
            req = yield from req

    @asyncio.coroutine
    def push_item(self, item):
            ret = self.sched.push_item(item)
            if isinstance(ret, asyncio.Future) or inspect.isgenerator(ret):
                ret = yield from ret

    def start_request(self):
        for spider in self.spiders.values():
            for req in spider.start_request():
                req.spider = spider.name
                yield from self.push_req(req)

    def run(self):
        yield from self.start_request()
        self.sched.start()

    def start(self, forever = True):
        self.loop.create_task(self.run())
        if forever:
            self.loop.run_forever()

    def shutdown(self):
        if self.loop.is_running():
            self.loop.stop()
        else:
            self.loop.close()



import asyncio
from .exceptions import IgnoreRequest, DropItem, ItemError

__all__ = ["BaseScheduler"]

class BaseScheduler(object):
    def __init__(self):
        self.engine = None
        self.is_running = False

    def push_req(self, req):
        '''
        push the request
        '''
        raise NotImplementedError('you must rewrite at sub class')

    def push_item(self, item):
        yield from self.submit_item(item)

    def submit_req(self, req):
        try:
            yield from self.engine.process(req)

        except IgnoreRequest:
            pass


    def submit_item(self, item):
        try:
            yield from self.engine.process_item(item)
        except (DropItem, ItemError):
            pass

    @asyncio.coroutine
    def run(self):
        '''
        run the scheduler
        '''
        raise NotImplementedError('you must rewrite at sub class')

    def start(self):
        if self.is_running:
            return

        self.is_running = True
        return self.engine.loop.create_task(self.run())



from importlib import import_module as _import_module
import logging

__all__ = ['import_module', 'import_pipelines', 'import_middlewares',
           'import_spiders', "logger"]

logger = logging.getLogger('grapy')

def import_module(module_name, *args, **kwargs):
    '''
    import the module and init it
    '''
    logger.debug('import module[%s]'%module_name)
    idx = module_name.rfind('.')
    module = _import_module(module_name[:idx])
    obj = getattr(module, module_name[idx+1:])
    return obj(*args, **kwargs)

def import_pipelines(pipelines):
    '''
    import module from a list like::

        [
            {'class_or_method:index': args},
            {'class_or_method:index': kwargs},
            {'class_or_method:index': None}
        ]

    the index is a number or string for order
    '''
    retval = []
    for module_name, values in pipelines.items():
        args = []
        kwargs = {}
        idx = module_name.find(':')
        order = 0
        if idx > -1:
            order = int(module_name[idx+1:])
            module_name = module_name[:idx]
        tp = type(values)
        if tp == tuple or tp == list:
            args = values
        elif tp == dict:
            keys = values.keys()
            if 'args' in keys or 'kwargs' in keys:
                args = values.get('args', ())
                kwargs = values.get('kwargs', {})
            else:
                kwargs = values

        elif values is not None:
            args.append(values)

        retval.append((import_module(module_name, *args, **kwargs), order))

    retval = [ret[0] for ret in sorted(retval, key=lambda x: x[1])]

    return retval

import_middlewares = import_pipelines
import_spiders = import_pipelines




__all__ = ['EngineError', 'DropItem', 'IgnoreRequest', 'RetryRequest', 'ItemError']

class EngineError(Exception):
    pass

class DropItem(Exception):
    pass

class IgnoreRequest(Exception):
    pass

class RetryRequest(Exception):
    pass

class ItemError(Exception):
    pass



import json
import re
import asyncio
import aiohttp
from ..utils import logger
from urllib.parse import urljoin
from .response import Response
from .exceptions import IgnoreRequest, RetryRequest
import requests
from time import time

__all__ = ['Request']

class Request(object):
    '''
    the Request object
    '''

    _keys = ['url', 'method', 'callback', 'callback_args', 'kwargs', 'spider',
            'req_id', 'group']
    _default = [{}, (), 'get', None, [], 'default']

    _json_keys = ['callback_args', 'kwargs']

    _null_char = '\x01'

    __slots__ = ['url', 'method', 'callback', 'callback_args', 'kwargs',
                 'spider', 'unique', 'req_id', 'ref', 'group', 'engine',
                 'request_time', 'http_proxy', 'async']

    def __init__(self, url, method='get',
            callback='parse', callback_args = [], **kwargs):
        self.url = re.sub('#.+', '', url)
        self.method = method
        self.callback = callback
        self.callback_args = callback_args
        self.kwargs = kwargs
        self.spider = 'default'
        self.unique = True
        self.req_id = 0
        self.ref = 0
        self.group = 0
        self.engine = None
        self.request_time = 0
        self.http_proxy = None
        self.async = True

    def pack(self):
        '''
        pack the Request object on bytes
        '''
        def _pack(key):
            val = getattr(self, key, '')
            if val not in self._default:
                if key in self._json_keys:
                    val = json.dumps(val)
            else:
                val = ''
            if not isinstance(val, str):
                val = str(val)
            return val
        return bytes(self._null_char.join(map(_pack, self._keys)), 'utf-8')

    def unpack(self, payload):
        '''
        unpack the Request payload
        '''
        payload = str(payload, 'utf-8')
        payload = payload.split(self._null_char)
        payload = dict(zip(self._keys, payload))

        for json_key in self._json_keys:
            if payload[json_key]:
                payload[json_key] = json.loads(payload[json_key])

        return payload

    def __bytes__(self):
        return self.pack()

    @classmethod
    def build(cls, payload):
        '''
        build a Request
        '''
        req = Request('')
        payload = req.unpack(payload)
        for key, val in payload.items():
            if val:
                if hasattr(req, key):
                    setattr(req, key, val)
        return req

    def _aio_request(self):
        method = self.method.lower()
        headers = self.engine.headers.copy()

        headers.update(self.kwargs.get('headers', {}))

        kwargs = {}

        kwargs.update(self.kwargs.copy())
        kwargs['headers'] = headers

        url = self.url

        start_time = time()

        with aiohttp.ClientSession(loop=self.engine.loop) as client:
            rsp = yield from client.request(method, url, **kwargs)

            ct = rsp.headers.get('content-type', '')
            logger.info('Request: {} {} {} {}'.format(method.upper(),
                                                      url, rsp.status, ct))
            if rsp.status >= 400 and rsp.status < 500:
                raise IgnoreRequest(url)
            if rsp.status == 200:
                if re.search('html|json|text|xml|rss', ct, re.I):
                    content = yield from rsp.read()
                    rsp.close()
                    self.request_time = time() - start_time
                    return Response(urljoin(url, str(rsp.url)), content, rsp)
                else:
                    raise IgnoreRequest(url)
            else:
                logger.error('Request fail: {} {}'.format(url, rsp.status))
                raise RetryRequest(url)

    def _request(self):
        url = self.url
        method = self.method.lower()
        headers = self.engine.headers.copy()
        headers.update(self.kwargs.get('headers', {}))
        kwargs = self.kwargs.copy()
        kwargs['headers'] = headers
        if self.http_proxy:
            kwargs['proxy'] = self.http_proxy,
        func = getattr(requests, method)
        rsp = func(url, **kwargs)
        ct = rsp.headers['content-type']
        if rsp.status_code >= 400 and rsp.status_code < 500:
            raise IgnoreRequest(url)
        if rsp.status_code == 200:
            if re.search('html|json|text|xml|rss', ct, re.I):
                return Response(urljoin(url, rsp.url), rsp.content, rsp)
            else:
                raise IgnoreRequest(url)
        else:
            logger.error('Request fail: {} {}'.format(url, rsp.status_code))
            raise RetryRequest(url)

    @asyncio.coroutine
    def request(self):
        '''
        do request default timeout is 300s

        >>> req = Request('http://example.com')
        >>> rsp = yield from req.request()
        '''
        start_time = time()

        try:
            if self.async:
                return (yield from self._aio_request())
            else:
                return self._request()
        except (aiohttp.http_exceptions.BadHttpMessage, aiohttp.http_exceptions.BadStatusLine, ValueError) as exc:
            logger.error(str(exc) + ': ' + self.url)
            start_time = time()
            return self._request()
        except aiohttp.client_exceptions.ClientError as e:
            logger.error("Request fail OsConnectionError: {} {}".format(self.url, e))
            raise IgnoreRequest(self.url)

        self.request_time = time() - start_time



from .request import Request

__all__ = ['BaseSpider']

class BaseSpider(object):
    '''The BaseSpider, all the spider recommend to extends this'''

    name = None
    start_urls = []

    def __init__(self, name=None, start_urls=[]):
        '''
        @name: the spider name, unique

        @start_urls: the start request url
        '''
        if not self.name:
            self.name = name
        if not self.start_urls:
            self.start_urls = start_urls

    def start_request(self):
        '''you can rewrite it for custem start request'''
        for url in self.start_urls:
            req = Request(url)
            req.unique = False
            yield req

    def parse(self, response):
        '''
        the default spider parse function.
        you must rewrite on a sub class.
        '''
        raise NotImplementedError('you must rewrite at sub class')



import json
import re
from .exceptions import ItemError
from ..utils import import_module
from uuid import uuid1 as uuid

__all__ = ['Item', 'load_item', 'dump_item']

class Item(object):
    _null_char = '\x01'

    _extra_field = {'name': 'extra', 'type': 'json'}

    _fields =  [
        {'name': 'extra', 'type': 'json'}
    ]

    __slots__ = ['__dict__']

    def __init__(self, payload = {}):

        if self._extra_field not in self._fields:
            self._fields.append(self._extra_field)

        if not isinstance(payload, dict):
            payload = self.unpack(payload)

        self.update(payload)

    def __getitem__(self, key, default=None):
        '''x.__getitem__(y) <==> x[y]'''
        return getattr(self, key, default)

    def __setitem__(self, key, val):
        '''x.__setitem__(i, y) <==> x[i]=y'''
        if isinstance(val, str):
            val = val.strip()
        setattr(self, key, val)

    def keys(self):
        '''D.keys() -> a set-like object providing a view on D's keys'''
        return self.__dict__.keys()

    def values(self):
        '''D.values() -> an object providing a view on D's values'''
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, key, default=None):
        '''
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        '''
        return self.__dict__.pop(key, default)

    def get(self, key, default=None):
        '''D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None.'''
        return self.__dict__.get(key, default)

    def update(self, item):
        '''
        D.update([E, ]**F) -> None.
        * Update D from dict/iterable E and F.
        * If E present and has a .keys() method, does:     for k in E: D[k] = E[k]
        * If E present and lacks .keys() method, does:     for (k, v) in E: D[k] = v
        * In either case, this is followed by: for k in F: D[k] = F[k]
        '''
        for k, v in item.items():
            if isinstance(v, str):
                item[k] = v.strip()
        return self.__dict__.update(item)

    def copy(self):
        return self.__dict__.copy()

    def pack(self):
        '''D.pack() -> a bytes object. pack item'''
        payload = dict(self)
        keys = list(map(lambda x: x['name'], self._fields))
        tps = list(map(lambda x: x['type'], self._fields))
        tps = dict(zip(keys, tps))

        none_keys = list(filter(lambda x: not payload[x], payload.keys()))
        list(map(payload.pop, none_keys))

        other_keys = filter(lambda x : x not in keys, payload.keys())
        other = dict(zip(other_keys, map(lambda x: payload[x], other_keys)))

        payload[self._extra_field['name']] = other

        def _pack(key):
            val = payload.get(key, '')
            tp = tps[key]

            if val:
                if tp == 'json':
                    val = json.dumps(val)
            else:
                val = ''
            if not isinstance(val, str):
                val = str(val)
            return val
        return self._null_char.join(map(_pack, keys))

    def unpack(self, payload):
        '''unpack item'''
        if isinstance(payload, bytes):
            payload = str(payload, 'utf-8')

        keys = list(map(lambda x: x['name'], self._fields))
        tps = list(map(lambda x: x['type'], self._fields))
        tps = dict(zip(keys, tps))

        payload = payload.split(self._null_char)

        def _unpack(pack):
            key, val = pack
            tp = tps[key]
            if not val:
                return key, val
            if tp == 'json':
                val = json.loads(val)
            elif tp == 'int':
                val = int(val)
            elif tp == 'float':
                val = float(val)
            elif tp == 'bool':
                val = bool(val)
            return key, val

        payload = dict(map(_unpack, zip(keys, payload)))

        if payload.get(self._extra_field['name']):
            other = payload.pop(self._extra_field['name'])
            if isinstance(other, dict):
                payload.update(other)

        return payload

    def __str__(self):
        return json.dumps(self.__dict__, indent=2)

    def __bytes__(self):
        return bytes(self.pack(), 'utf-8')

    @property
    def unique(self):
        return str(uuid())

NULL_CHAR = '\x02\x00\x00'
def dump_item(klass, *args, **kwargs):
    '''dump the Item'''
    cls = klass.__class__
    cls_name = re.search("'([^']+)'", str(cls)).group(1)
    if not isinstance(klass, Item):
        raise ItemError(
                'ItemError: %s is not instance crawl.core.item.Item'%cls_name)
    retval = NULL_CHAR.join([cls_name, klass.pack()])
    return retval

def load_item(string):
    '''load the Item'''
    cls_name, data = string.split(NULL_CHAR)
    klass = import_module(cls_name, data)
    if not isinstance(klass, Item):
        raise ItemError(
                'ItemError: %s is not instance crawl.core.item.Item'%cls_name)
    return klass



import re
from bs4 import BeautifulSoup
import json

RE_XML = re.compile('<\?xml.+encoding=["\']([^\'"]+?)["\'].+\?>', re.I)
RE_HTML = re.compile('<meta.+charset=["\']([^\'"]+?)[\'"].+>', re.I)

__all__ = ['Response']

class Response(object):

    __slots__ = ['url', 'raw', 'encoding', 'content', '_soup', 'req', 'headers']

    def __init__(self, url, content, raw):
        self.raw = raw
        self.url = url
        self._soup = None
        self.encoding = None
        self.content = content
        self.req = None
        self.headers = raw.headers

    @property
    def text(self):
        'return the unicode document'
        content = self.content
        if self.encoding:
            return str(content, self.encoding, errors = 'ignore')

        charset = self._get_charset(content)
        if charset:
            self.encoding = charset
            return str(content, charset, errors = 'ignore')
        else:
            try:
                self.encoding = 'GBK'
                return str(content, 'GBK')
            except UnicodeDecodeError:
                self.encoding = 'UTF-8'
                return str(content, 'UTF-8', errors = 'ignore')

    def json(self):
        '''return json document, maybe raise'''
        data = self.content
        data = json.loads(data.decode('utf-8'))
        return data


    @property
    def soup(self):
        '''return the instance of BeautifulSoup'''
        if self._soup is None:
            text = self.text
            self._soup = BeautifulSoup(text, 'html.parser')
        return self._soup

    def _get_charset(self, content):

        def map_charset(charset):
            if charset:
                charset = charset.upper()
                if charset == 'GB2312':
                    charset = 'GBK'
            return charset

        ct = ''
        try:
            ct = self.headers.get('content-type', '').lower()
        except:
            pass
        p = re.search('charset=(.+)$', ct)
        if p:
            charset = p.group(1)
            return map_charset(charset)

        content = str(content, 'utf-8', errors='ignore')
        xml = RE_XML.search(content)
        if xml:
            charset = xml.group(1)
            return map_charset(charset)

        html = RE_HTML.search(content)
        if html:
            charset = html.group(1)
            return map_charset(charset)

        return None

    def select(self, selector):
        '''
        select elements use the css selector
        '''
        soup = self.soup
        re_tag = re.compile('^[a-z0-9]+$', re.I | re.U)
        re_attribute = re.compile('^(?P<tag>\w+)?\[(?P<attribute>[a-z\-_]+)(?P<operator>[=~\|\^\$\*]?)=?"?(?P<value>[^\]"]*)"?\]$')

        def attribute_checker(operator, attribute, value = ''):
            """
            Takes an operator, attribute and optional value; returns a function
            that will return True for elements that match that combination.
            """

            return {
                '=': lambda el: el.get(attribute) == value,
                # attribute includes value as one of a set of space separated tokens
                '~': lambda el: value in el.get(attribute, '').split(),
                # attribute starts with value
                '^': lambda el: el.get(attribute, '').startswith(value),
                # attribute ends with value
                '$': lambda el: el.get(attribute, '').endswith(value),
                # attribute contains value
                '*': lambda el: value in el.get(attribute, ''),
                # attribute is either exactly value or starts with value-
                '|': lambda el: el.get(attribute, '') == value \
                        or el.get(attribute, '').startswith('%s-' % value),
            }.get(operator, lambda el: el.has_attr(attribute))

        tokens = selector.split()
        current_context = [soup]

        for index, token in enumerate(tokens):
            if tokens[index - 1] == '>':
                continue

            m = re_attribute.match(token)
            if m:
                # Attribute selector
                tag, attribute, operator, value = m.groups()

                if not tag:
                    tag = True

                checker = attribute_checker(operator, attribute, value)

                found = []
                for context in current_context:
                    found.extend([el for el in context.find_all(tag) if checker(el)])

                current_context = found
                continue

            if '#' in token:
                # ID selector
                tag, id = token.split('#', 1)
                if not tag:
                    tag = True

                el = current_context[0].find(tag, {'id': id})
                if not el:
                    return []

                current_context = [el]
                continue

            if '.' in token:
                # Class selector
                tag, klass = token.split('.', 1)
                if not tag:
                    tag = True

                klasses = set(klass.split('.'))
                found = []
                for context in current_context:
                    found.extend(
                        context.find_all(tag, {'class': lambda attr:
                            attr and klasses.issubset(attr.split())})
                    )

                current_context = found
                continue

            if '*' in token:
                # Star selector
                found = []
                for context in current_context:
                    found.extend(context.find_all(True))

                current_context = found
                continue

            if token == '>':
                # Child selector
                tag = tokens[index + 1]
                if not tag:
                    tag = True

                found = []
                for context in current_context:
                    found.extend(context.find_all(tag, recursive=False))

                current_context = found
                continue

            # Here we should just have a regular tag
            if not re_tag.match(token):
                return []

            found = []
            for context in current_context:
                found.extend(context.find_all(token))

            current_context = found

        return current_context



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

packages = [
    'grapy',
    'grapy.core',
]

requires = ['asyncio', 'aiohttp', 'beautifulsoup4', 'requests']

setup(
    name='grapy',
    version='0.1.8',
    description='Grapy, a fast high-level screen scraping and web crawling framework for Python 3.3 or later base on asyncio.',
    author='Li Meng Jun',
    author_email='lmjubuntu@gmail.com',
    url='https://github.com/Lupino/grapy',
    packages=packages,
    package_dir={'grapy': 'grapy'},
    include_package_data=True,
    install_requires=requires,
)



from .core import BaseScheduler
import hashlib
import asyncio
import re
from .utils import logger
re_url = re.compile('^https?://[^/]+')

__all__ = ['Scheduler']

def hash_url(url):
    h = hashlib.sha1()
    h.update(bytes(url, 'utf-8'))
    return h.hexdigest()

class Scheduler(BaseScheduler):
    def __init__(self, storage = {}, queue=[], max_tasks=5):
        BaseScheduler.__init__(self)
        self._storage = storage
        self._queue = queue
        self._sem = asyncio.Semaphore(max_tasks)

    def push_req(self, req):
        if not re_url.match(req.url):
            return
        key = hash_url(req.url)
        if req.unique and key in self._storage:
            return

        self._queue.insert(0, req)
        self._storage[key] = {'key': key, 'req': req, 'crawled': False}

        self.start()

    def run(self):
        while True:
            if len(self._queue) == 0:
                break

            req = self._queue.pop()
            yield from self._sem.acquire()
            task = self.engine.loop.create_task(self.submit_req(req))
            task.add_done_callback(lambda t: self._sem.release())

        self.is_running = False

    def submit_req(self, req):
        try:
            yield from BaseScheduler.submit_req(self, req)
        except Exception as e:
            logger.exception(e)
        key = hash_url(req.url)
        self._storage[key] = {'key': key, 'req': req, 'crawled': True}

    def submit_item(self, item):
        try:
            yield from BaseScheduler.submit_item(self, item)
        except Exception as e:
            logger.exception(e)



from .engine import Engine
from .base_spider import BaseSpider
from .base_sched import BaseScheduler
from .request import Request
from .response import Response
from .item import Item, dump_item, load_item

__all__ = ['Engine', 'BaseSpider', 'BaseScheduler', 'Request', 'Response',
    'Item', 'dump_item', 'load_item']
